/*********************************************************  DMA Driver  *********************************************************
	Red Rapids
	797 N. Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	Description: DMA buffer driver for Red Rapids PCI hardware.

	Rev History
		02/19/2020	-	Modified the call to dma_mmap_coherent() to pass the device structure instead of NULL in first argument.
		08/08/2018	-	Initial release.
		
	Copyright (c) 2020  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*********************************************************  DMA Driver  *********************************************************/

#include <linux/cdev.h>
#include <linux/dma-mapping.h>
#include <linux/kdev_t.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/pci.h>

//////////////////////////////////////////////////  DMA BUFFER CONFIGURATION  //////////////////////////////////////////////////

#define DMA_DEV_NAME "rrdma0"		// Name of the character device that will appear in /dev.
#define NUMBER_OF_BUFFERS 8			// Number of separate DMA buffers created by this driver (min=1, max=100).
#define DFLT_PAGE_COUNT 2			// Default number of DMA pages per buffer.
#define DFLT_PAGES_PER_MARK 1		// Default number of DMA pages to transfer between completion marker flags.
#define DFLT_BURST_SIZEB 4096		// Default size (bytes) of each DMA burst transaction.
#define DFLT_BURST_COUNT 2			// Default number of burst transactions per DMA page.

// Assign the default values to the configuration variables of each buffer.
// Defaults can be substituted with other values, for example: dma_page_count[NUMBER_OF_BUFFERS] = {4, 100, 15, 1024}
size_t dma_burst_sizeB[NUMBER_OF_BUFFERS] 	 = 	{DFLT_BURST_SIZEB, DFLT_BURST_SIZEB, DFLT_BURST_SIZEB, DFLT_BURST_SIZEB,
												DFLT_BURST_SIZEB, DFLT_BURST_SIZEB, DFLT_BURST_SIZEB, DFLT_BURST_SIZEB};
size_t dma_burst_count[NUMBER_OF_BUFFERS] 	 = 	{DFLT_BURST_COUNT, DFLT_BURST_COUNT, DFLT_BURST_COUNT, DFLT_BURST_COUNT,
												DFLT_BURST_COUNT, DFLT_BURST_COUNT, DFLT_BURST_COUNT, DFLT_BURST_COUNT};
size_t dma_page_count[NUMBER_OF_BUFFERS] 	 = 	{DFLT_PAGE_COUNT, DFLT_PAGE_COUNT, DFLT_PAGE_COUNT, DFLT_PAGE_COUNT,
												DFLT_PAGE_COUNT, DFLT_PAGE_COUNT, DFLT_PAGE_COUNT, DFLT_PAGE_COUNT};
size_t dma_pages_per_mark[NUMBER_OF_BUFFERS] =	{DFLT_PAGES_PER_MARK, DFLT_PAGES_PER_MARK, DFLT_PAGES_PER_MARK, DFLT_PAGES_PER_MARK,
												DFLT_PAGES_PER_MARK, DFLT_PAGES_PER_MARK, DFLT_PAGES_PER_MARK, DFLT_PAGES_PER_MARK}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define RR_PCI_VENDOR_ID 0x17D2					// Red Rapids PCI vendor ID assigned by PCI-SIG.
#define RR_PCI_DEVICE_ID 0x00CA					// Red Rapids device ID assigned to each product.

struct cdev dmabuf_cdev;						// Character device structure used by the kernel.
dev_t dmabuf_dev;								// Character device numbers.
dma_addr_t *dmabuf_handle[NUMBER_OF_BUFFERS];	// Physical address of allocated contiguous DMA pages for each device.
struct class *p_class;							// Pointer to a struct class.
struct device *p_device[NUMBER_OF_BUFFERS];		// Pointer to a struct device.
void **p_kva[NUMBER_OF_BUFFERS];				// Pointer to the kernel virtual address of allocated contiguous DMA pages for each device.
size_t dma_page_index[NUMBER_OF_BUFFERS];		// Index for user space to sequentially access each of the DMA pages allocated to a buffer.
size_t dma_page_size[NUMBER_OF_BUFFERS];		// Number of DMA pages created for each buffer.

// Extract the buffer index from the character file name.
static size_t get_number(char *filename)
{
	char *dash;
	
	// Find the "-" character that precedes the DMA buffer index.
	dash = strchr(filename, '-');
	if (strlen(dash) - 2)
		// Return a value between ten and ninety-nine.
		return (10 * (dash[1] - 48)) + dash[2] - 48;
	else
		// Return a value less than ten.
		return dash[1] - 48;
}
	
// Open the character device.
static int dmabuf_open(struct inode *inode, struct file *p_file)
{
	unsigned long long *p_dma_page;
	size_t chan_index;		// Channel index making the request.
	size_t dma_page_cnt;	// DMA page counter.
	
	// Extract the DMA buffer index number from the character file name.
	chan_index = get_number(p_file->f_path.dentry->d_iname);
	
	printk(KERN_INFO "%s: buffer index #%u opened\n", DMA_DEV_NAME, (unsigned)chan_index);
	
	// Write a page descriptor into the start of each DMA page that can be read in user space.
	for (dma_page_cnt = 0; dma_page_cnt < dma_page_count[chan_index]; dma_page_cnt++)
	{
		p_dma_page = (unsigned long long *)p_kva[chan_index][dma_page_cnt];
		p_dma_page[0] = (unsigned long long)dma_page_cnt;								// Sequential page number.
		p_dma_page[1] = (unsigned long long)dma_page_count[chan_index];					// Number of pages available.
		p_dma_page[2] = (unsigned long long)dma_pages_per_mark[chan_index];				// Number of pages between markers.
		p_dma_page[3] = (unsigned long long)dma_burst_sizeB[chan_index];				// Burst size.
		p_dma_page[4] = (unsigned long long)dma_burst_count[chan_index];				// Burst count.
		p_dma_page[5] = (unsigned long long)dmabuf_handle[chan_index][dma_page_cnt];	// Physical address of page.
	}
	
	// Reset the DMA page index.
	dma_page_index[chan_index] = 0;
	
	return 0;
}

// Release the character device.
static int dmabuf_release(struct inode *inode, struct file *p_file)
{
	size_t chan_index;		// Channel index making the request.
	
	// Extract the DMA buffer index number from the character file name.
	chan_index = get_number(p_file->f_path.dentry->d_iname);
	
	printk(KERN_INFO "%s: buffer index #%u released\n", DMA_DEV_NAME, (unsigned)chan_index);

	return 0;
}

// Map a user space pointer to the allocated DMA buffer.
static int dmabuf_mmap(struct file *p_file, struct vm_area_struct *p_vma)
{
	int rc = 0;				// Function return code.
	size_t chan_index;		// Channel number making the request.
	
	// Extract the DMA buffer index from the character file name.
	chan_index = get_number(p_file->f_path.dentry->d_iname);
	
	// Return an error if the number of bytes requested is greater than the number allocated.
	if ((p_vma->vm_end - p_vma->vm_start) > dma_page_size[chan_index])
		return -EIO;
	
	// Return the user space pointer or a negative error code.
	rc = dma_mmap_coherent(p_device[chan_index], p_vma, p_kva[chan_index][dma_page_index[chan_index]], dmabuf_handle[chan_index][dma_page_index[chan_index]], p_vma->vm_end - p_vma->vm_start);
	
	// Increment the DMA page index.
	if (dma_page_index[chan_index] == dma_page_count[chan_index]-1)
		dma_page_index[chan_index] = 0;
	else
		dma_page_index[chan_index]++;
	
	return rc;
}

// Assign functions to the file operations data structure.
static struct file_operations dmabuf_fops = {
        .open = dmabuf_open,
        .release = dmabuf_release,
        .mmap = dmabuf_mmap,
        .owner = THIS_MODULE,
};

// Initialization function that runs when the kernel module is loaded.
static int __init dmabuf_init(void)
{
	int rc = 0;					// Function return code.
	size_t dma_chan_cnt;		// DMA buffer counter.
	size_t dma_page_cnt;		// DMA page counter.
	struct pci_dev *p_pci_dev;
	
	// Scan for a Red Rapids device.
	if ((p_pci_dev = pci_get_device(RR_PCI_VENDOR_ID, RR_PCI_DEVICE_ID, NULL)) == NULL)
	{
		printk(KERN_ERR "%s: Red Rapids device not found\n", DMA_DEV_NAME);
		return -ENXIO;
	}

	// Initialize variables.
	for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
	{
		// Set the page index to zero (first page) for each buffer
		dma_page_index[dma_chan_cnt] = 0;

		// Initialize pointers to assist with clean-up on error detection.
		p_device[dma_chan_cnt] = NULL;
		p_kva[dma_chan_cnt] = NULL;
		dmabuf_handle[dma_chan_cnt] = NULL;
		
		// Calculate the requested DMA page size in bytes.
		dma_page_size[dma_chan_cnt] = dma_burst_sizeB[dma_chan_cnt] * dma_burst_count[dma_chan_cnt];
		// Increase the requested DMA page size to the next Linux page size boundary:
		dma_page_size[dma_chan_cnt] += dma_page_size[dma_chan_cnt] % PAGE_SIZE;
	}

	// Allocate variable storage space to each buffer.
	for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
	{
		// Allocate memory to store the DMA handles.
		dmabuf_handle[dma_chan_cnt] = kzalloc(dma_page_count[dma_chan_cnt] * sizeof(dma_addr_t), GFP_KERNEL);
		if (!dmabuf_handle[dma_chan_cnt])
		{
			printk(KERN_ERR "%s: handle memory allocation failed\n", DMA_DEV_NAME);
			rc = -ENOMEM;
			goto init_error1;
		}
	
		// Allocate memory to store the virtual address pointers.
		p_kva[dma_chan_cnt] = kzalloc(dma_page_count[dma_chan_cnt] * sizeof(void *), GFP_KERNEL);
		if (!p_kva[dma_chan_cnt])
		{
			printk(KERN_ERR "%s: virtual address pointer memory allocation failed\n", DMA_DEV_NAME);
			rc = -ENOMEM;
			goto init_error1;
		}
		
		// Initialize pointers to assist with clean-up on error detection.
		for (dma_page_cnt = 0; dma_page_cnt < dma_page_count[dma_chan_cnt]; dma_page_cnt++)
			p_kva[dma_chan_cnt][dma_page_cnt] = NULL;
	}
	
	// Allow 64-bit DMA addresses.
	if(dma_set_coherent_mask(&p_pci_dev->dev, DMA_BIT_MASK(64)))
		printk(KERN_WARNING "%s: unable to set 64-bit DMA addressing\n", DMA_DEV_NAME);

	// Allocate the requested number of DMA pages to each buffer.
	for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
		// Allocate the requested number of contiguous DMA pages.
		for (dma_page_cnt = 0; dma_page_cnt < dma_page_count[dma_chan_cnt]; dma_page_cnt++)
		{
			p_kva[dma_chan_cnt][dma_page_cnt] = dma_alloc_coherent(&p_pci_dev->dev, dma_page_size[dma_chan_cnt], &dmabuf_handle[dma_chan_cnt][dma_page_cnt], GFP_KERNEL);
			if (!p_kva[dma_chan_cnt][dma_page_cnt])
			{
				printk(KERN_ERR "%s: contiguous memory allocation failed\n", DMA_DEV_NAME);
				rc = -ENOMEM;
				goto init_error2;
			}
		}

	// Allocate a character device number for this driver.
	if ((rc = alloc_chrdev_region(&dmabuf_dev, 0, NUMBER_OF_BUFFERS, DMA_DEV_NAME)) < 0)
	{
		printk(KERN_ERR "%s: could not allocate character device number\n", DMA_DEV_NAME);
		goto init_error2;
	}

	// Initialize the device structure and register the device with the kernel.
	cdev_init(&dmabuf_cdev, &dmabuf_fops);
	if ((rc = cdev_add(&dmabuf_cdev, dmabuf_dev, NUMBER_OF_BUFFERS)) < 0)
	{
       	printk(KERN_ERR "%s: could not add character device\n", DMA_DEV_NAME);
        goto init_error3;
	}

	// Create a struct class pointer.
	if (IS_ERR(p_class = class_create(THIS_MODULE, DMA_DEV_NAME)))
	{
		printk(KERN_ERR "%s: could not create class\n", DMA_DEV_NAME);
       	rc = PTR_ERR(p_class);
        goto init_error4;
	}
        
	// Create the requested number of devices and register them with sysfs.
	for (dma_page_cnt = 0; dma_page_cnt < NUMBER_OF_BUFFERS; dma_page_cnt++)
		if (IS_ERR(p_device[dma_page_cnt] = device_create(p_class, NULL, MKDEV(MAJOR(dmabuf_dev), MINOR(dmabuf_dev) + dma_page_cnt), NULL, "%s-%zu", DMA_DEV_NAME, dma_page_cnt)))
		{
			printk(KERN_ERR "%s: could not create device\n", DMA_DEV_NAME);
			rc = PTR_ERR(p_device[dma_page_cnt]);
			goto init_error5;
		}
		
	return 0;

	// Clean up the current state if an error is detected during initialization.
    init_error5:
    	for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
    		if (!p_device[dma_chan_cnt] && !IS_ERR(p_device[dma_chan_cnt]))
    			device_destroy(p_class, MKDEV(MAJOR(dmabuf_dev), MINOR(dmabuf_dev) + dma_chan_cnt));
    	class_destroy(p_class);
    init_error4:
    	cdev_del(&dmabuf_cdev);
	init_error3:
		unregister_chrdev_region(dmabuf_dev, 1);
	init_error2:
		for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
			for (dma_page_cnt = 0; dma_page_cnt < dma_page_count[dma_chan_cnt]; dma_page_cnt++)
				if (!p_kva[dma_chan_cnt][dma_page_cnt])
					dma_free_coherent(NULL, dma_page_size[dma_chan_cnt], p_kva[dma_chan_cnt][dma_page_cnt], dmabuf_handle[dma_chan_cnt][dma_page_cnt]);
	init_error1:
		for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
		{
			if (!dmabuf_handle[dma_chan_cnt])
				kfree(dmabuf_handle[dma_chan_cnt]);
			if (!p_kva[dma_chan_cnt])
				kfree(p_kva[dma_chan_cnt]);
		}
	return rc;
}

// Exit function that runs when the kernel module is unloaded (rmmod).
static void __exit dmabuf_exit(void)
{
	size_t dma_chan_cnt;	// DMA buffer counter.
	size_t dma_page_cnt;	// DMA page counter.
	
	// Remove the character device.
    for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
    	device_destroy(p_class, MKDEV(MAJOR(dmabuf_dev), MINOR(dmabuf_dev) + dma_chan_cnt));
    class_destroy(p_class);
    cdev_del(&dmabuf_cdev);
	unregister_chrdev_region(dmabuf_dev, 1);

	// Free allocated memory.
	for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
	{
		for (dma_page_cnt = 0; dma_page_cnt < dma_page_count[dma_chan_cnt]; dma_page_cnt++)
			dma_free_coherent(NULL, dma_page_size[dma_chan_cnt], p_kva[dma_chan_cnt][dma_page_cnt], dmabuf_handle[dma_chan_cnt][dma_page_cnt]);
		kfree(p_kva[dma_chan_cnt]);
	}
	for (dma_chan_cnt = 0; dma_chan_cnt < NUMBER_OF_BUFFERS; dma_chan_cnt++)
		kfree(dmabuf_handle[dma_chan_cnt]);
}

module_init(dmabuf_init);
module_exit(dmabuf_exit);
MODULE_DESCRIPTION("DMA Buffer Allocation");
MODULE_AUTHOR("Red Rapids");
MODULE_LICENSE("GPL");