/**********************************************************  SigFPGA  *********************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following code demonstrates the operation of all Red Rapids SigFPGA PCI Express products.

	Note: The application can be applied to the maximum number of Red Rapids devices supported by the host.

	Rev History:  01-20-2020:	Removed double device counter loop around the final FreeResources() function.
				  				Added DevNum assignment to the hardware handle so multiple devices are recognized.
								Added SigFPGAID() to uniquely identify the model number when multiple devices are present.
								Moved parenthesis in call to MemTest().
				  12-09-2019:	Moved SoftwareReset() to HWInitialize() since it is synchronous and relies on a stable clock.
				  03-11-2019:	Some function and variable names changed for consistency across products.
				  11-28-2018:	Style and format changes.
				  08-08-2018:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

**********************************************************  SigFPGA  *********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>

#include "driver_functions.h"			// Functions to communicate with the UIO and DMA drivers.
#include "sigfpga.h"					// SigFPGA demonstration application.
#include "chip_functions.h"				// Functions to program specific chips used in the SigFPGA design.
#include "product_functions.h"			// Functions that apply to specific SigFPGA products.
#include "global_functions.h"			// Functions that apply to all SiGFPGA hardware and channel types.
#include "channel_functions.h"			// Functions that apply to individual SiGFPGA channels.
#include "dma_functions.h"				// Functions to manage DMA transactions with the host.
#include "utility_functions.h"			// Functions specific to this application example.
#include "default_hardware_config.h"	// Default hardware configuration values.
#include "default_operating_config.h"	// Default operating configuration values.

int main()
{
	s_UIOScan UIO[NUM_DEVICES];			// Descriptor for each UIO device identified as a Red Rapids product.
	s_Handle *p_HW[NUM_DEVICES];		// Hardware handle to each Red Rapids device.
	s_User User[NUM_DEVICES];			// User defined structure attached to the hardware handle that can be customized.
	unsigned channel_count[NUM_DEVICES] = {0};	// Total number of RX and TX channels available in each device.
	int status = 0;						// Variable to store status returned by function calls.
	unsigned index = 0;					// Variable to index array elements in loops.
	unsigned devcnt = 0;				// Variable to index device numbers in loops.
	unsigned devfound = 0;				// Number of Red Rapids devices detected.
	unsigned devsigfpga = 0;			// Number of Red Rapids SigFPGA devices detected.
	unsigned chancnt = 0;				// Variable to index channel numbers in loops.
	unsigned pagecount = 0;				// Variable to index DMA page numbers in loops.
	unsigned model = 0;					// Model number reported by a Red Rapids device.
	unsigned error[4] = {0};			// Variable to record detected configuration errors.
	s_ErrorStatus Error;				// Structure to store the results returned by a fault status query to the device.
	s_Temperature Temperature;			// Structure to store the results returned by the temperature monitor on the device.
	s_Power Power;						// Structure to store the results returned by the power monitor on the device.
	float performance[4] = {0};			// Array to store PCI RX throughput (0), TX throughput (1), RX latency (2) and TX latency (3) measurements.
	double fsdivisor = 8.0;				// Variable to set the transmitter output frequency for demonstration (Freq = Fs / FsDivisor).
	char filename[16];					// File name to store receiver dma buffer contents and data file contents.
	
	printf("******************************** SigFPGA ********************************\n");

	///////////////////////////////////////////  Establish UIO Driver Communication  ///////////////////////////////////////////

	printf("\nScanning for Red Rapids hardware...\n");
	
	// Scan the UIO devices for Red Rapids devices.
	devfound = UIOScan(UIO);
	if (devfound == 0 || devfound > NUM_DEVICES)
	{
		if (devfound == 0) printf("\n...terminating execution because no Red Rapids device found.\n");
		if (devfound > NUM_DEVICES) printf("\n...terminating execution because more Red Rapids devices found than expected.\n");
		return EXIT_FAILURE;
	}
	
	// Initialize the hardware handle for each device.
	for (devcnt = 0; devcnt < devfound; devcnt++)
		// Allocate memory for the structure and initialize if successful.
		if ((p_HW[devsigfpga] = malloc(sizeof *p_HW[devsigfpga])) == NULL)
		{
			printf("\tERROR: Unable to allocate memory for the device #%u hardware handle.\n", devcnt);
			// Free any allocated memory.
			if (devsigfpga)
				for (index = 0; index < devsigfpga; index++)
					free(p_HW[index]);
			printf("\n...terminating execution due to hardware handle memory allocation error.\n");
			return EXIT_FAILURE;
		}
		else
		{
			// Initialize each member of the device handle structure.
			DEVInit(p_HW[devsigfpga], devcnt);
			// Set a unique device number for each hardware handle.
			p_HW[devsigfpga]->DevNum = devcnt;
			if (!SigFPGAID(UIO[devsigfpga], p_HW[devsigfpga]))
			{
				// Attach the User structure created for this application to the device hardware handle.
				UserInit(p_HW[devsigfpga], &User[devcnt]);
				devsigfpga++;
			}
			else
				// Free the allocated memory since this is not a SigFPGA device.
				free(p_HW[devsigfpga]);
		}

	// Exit if no SigFPGA devices were detected.
	if (!devsigfpga)
		return EXIT_FAILURE;

	//////////////////////////////////////////////////  Initialize Hardware  //////////////////////////////////////////////////

	// Initialize the hardware settings for each Red Rapids device.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
	{
		printf("\nInitializing device DEV%u...\n", devcnt);

		// Open the UIO character device file assigned to each Red Rapids device.
		if ((status = DEVOpen(UIO[devcnt], &p_HW[devcnt]->DEV)))
			printf("\t[ERROR] UIO character device could not be opened.\n");
		else
		{
			// Display the firmware revision date.
			printf("\t[INFO] Firmware revision (MMDDYYYY): %08X\n", FirmwareRevision(p_HW[devcnt]));

			// Report the model number returned from the Red Rapids device.
			model = ModelNumber(p_HW[devcnt]);
			printf("\t[INFO] Identified Red Rapids Model %03X.\n", model);
			
			// Peform model number specific hardware configuration procedures.
			if ((status = HWInitialize(p_HW[devcnt], model)))
				printf("\t[ERROR] Model %03X hardware initialization failed.\n", model);
			
			// Report the clock status returned from the Red Rapids device.
			switch (ClockStatus(p_HW[devcnt]))
			{
				case 0: {printf("\t[INFO] External sample clock active, converter clocks locked.\n"); break;}
				case 1: {printf("\t[ERROR] External sample clock active, converter clocks not locked.\n"); status = CLOCK_LOCK; break;}
				case 2: {printf("\t[INFO] Sample clock locked to internal reference, converter clocks locked.\n"); break;}
				case 3: {printf("\t[ERROR] Sample clock locked to internal reference, converter clocks not locked.\n"); status = CLOCK_LOCK; break;}
				case 4: {printf("\t[INFO] Sample clock locked to external reference, converter clocks locked.\n"); break;}
				case 5: {printf("\t[ERROR] Sample clock locked to external reference, converter clocks not locked.\n"); status = CLOCK_LOCK; break;}
				default: {printf("\t[ERROR] Sample clock is not locked.\n"); status = SYNTH_LOCK; break;}
			}
			
			if (!status)
			{
				// Calculate the total number of RX and TX channels available in the device.
				channel_count[devcnt] = User[devcnt].RXChannelCount + User[devcnt].TXChannelCount;
			
				// Allocate memory to store error flags for each channel.
				if ((User[devcnt].Error = malloc(sizeof *User[devcnt].Error * channel_count[devcnt])) == NULL)
					status = errno;
				else
					for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
						if ((User[devcnt].Error[chancnt-1] = malloc(sizeof *User[devcnt].Error[chancnt-1] * 4)) == NULL)
							status = errno;
				
				if (status)
					printf("\t[ERROR] Memory allocation for channel variables failed.\n");
			}
		}
		
		// Exit if any errors were detected.
		if (status)
		{
			// Free allocated resources.
			for (index = 0; index < devsigfpga; index++)
				FreeResources(p_HW[index], channel_count[index], UIO[index]);
			printf("\n...terminating execution due to hardware initialization errors.\n");
			return EXIT_FAILURE;
		}
	}
	
	///////////////////////////////////////////////////  Assign DMA Buffers  ///////////////////////////////////////////////////
	
	// Memory map the DMA buffer assigned to each Red Rapids device.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
	{
		printf("\nAssigning DMA buffers to device DEV%u...\n", devcnt);
		
		// Allocate memory for the DMA structure attached to the hardware handle.
		if ((p_HW[devcnt]->DMA = malloc(sizeof *p_HW[devcnt]->DMA * channel_count[devcnt])) == NULL)
			status = errno;
		else
			// Initialize each member of the DMA structure.
			for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
				DMAInit(&p_HW[devcnt]->DMA[chancnt-1]);
		
		// Allocate memory for the User data buffer that is used to demonstrate the transfer of data with DMA buffers.
		if ((User[devcnt].p_DataBuffer = malloc(sizeof *User[devcnt].p_DataBuffer * channel_count[devcnt])) == NULL)
			status = errno;
		
		// Do not proceed if any of the memory allocations failed.
		if (status)
			printf("\t[ERROR] Memory allocation for DMA variables failed.\n");
		else
			for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
			{
				// Open the DMA buffer character device file and retrieve configuration information.
				// Note: The DMA character device names (rrdma#-#) correspond to Red Rapids device numbers (first #) followed by a channel index (second #).
				if ((status = DMAOpen(&p_HW[devcnt]->DMA[chancnt-1], devcnt, chancnt - 1)))
					printf("\t[ERROR] Channel #%u DMA character device could not be opened.\n",chancnt);
				else
				{
					// Allocate enough memory to allow the User buffer to mirror the data stored in all of the DMA pages between markers for every channel.
					if ((User->p_DataBuffer[chancnt-1] = malloc(sizeof *User->p_DataBuffer[chancnt-1] * p_HW[devcnt]->DMA[chancnt-1].PagesPerMark)) == NULL)
						status = errno;
					else
						for (pagecount = 0; pagecount < p_HW[devcnt]->DMA[chancnt-1].PagesPerMark; pagecount++)
							if ((User->p_DataBuffer[chancnt-1][pagecount] = malloc(p_HW[devcnt]->DMA[chancnt-1].BurstSizeB * p_HW[devcnt]->DMA[chancnt-1].BurstCount)) == NULL)
								status = errno;
			
					// Do not proceed if any of the memory allocations failed.
					if (status)
						printf("\t[ERROR] Memory allocation for User data buffer failed.\n");
					else
					{
						// Check the DMA configuration settings for possible errors.
						if((status = DMAConfigCheck(p_HW[devcnt]->DMA[chancnt-1], error)))
						{
							// This is a fatal error since the DMA buffers cannot be correctly configured.
							if (error[0]) printf("\t[ERROR] Channel #%u DMA buffer BurstSizeB exceeds maximum value or illegal alignment.\n", chancnt);
							if (error[1]) printf("\t[ERROR] Channel #%u DMA buffer BurstCount exceeds maximum value.\n", chancnt);
							if (error[2]) printf("\t[ERROR] Channel #%u DMA buffer PageCount exceeds maximum value.\n", chancnt);
							if (error[3]) printf("\t[ERROR] Channel #%u DMA buffer PagesPerMark not an even multiple of PageCount.\n", chancnt);
						}
						else
							// Assign DMA descriptors and physical page addresses to each device channel.
							DMAAssign(p_HW[devcnt], chancnt - 1);
					}
				}
				
				// Free memory allocated by DMAOpen() since the DMA physical addresses are no longer needed.
				free(p_HW[devcnt]->DMA[chancnt-1].Address);
			
			// Exit if any errors were detected.
			if (status)
			{
				// Free allocated resources.
				for (index = 0; index < devsigfpga; index++)
					FreeResources(p_HW[index], channel_count[index], UIO[index]);
				printf("\n...terminating execution due to DMA buffer assignment errors.\n");
				return EXIT_FAILURE;
			}
		}
	}

	//////////////////////////////////////////////  Load Configuration Settings  //////////////////////////////////////////////

	// Load the operating configuration settings to each Red Rapids device.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
	{
		printf("\nLoading device DEV%u configuration settings...\n", devcnt);

		// Set the direction of the GPIO and USER connector pins.
		// Note: The GPIOConnect() and/or USERConnect() functions must be called to enable each bus.
		PinDirection(p_HW[devcnt], PinDirectionDefault);
		// Set the interrupt mask bits for individual fault conditions.
		ErrorMask(p_HW[devcnt], ErrorMaskDefault);
		// Set the interrupt mask bits for individual channel status events.
		for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
			ChannelMask(p_HW[devcnt], chancnt, ChannelMaskDefault);
		// Set the global interrupt mask.
		InterruptMask(p_HW[devcnt], GlobalMaskDefault);

		// Configure each channel with application specific settings.
		for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
		{
			// Load the synchronizer settings that control channel activity.
			SynchronizerSettings(p_HW[devcnt], chancnt, SynchronizerDefault);
			// Transmitter channels must be preloaded with data before the initial DMA transfer is requested.
			if (chancnt > User[devcnt].RXChannelCount)
			{
				// Create a file of sine wave samples to demonstrate transmitter channels (Frequency = Fs / FsDivisor).
				sprintf(filename, "txdata%u-%u.txt", devcnt, chancnt);
				// Generate enough points to fill one complete DMA page with 16-bit data items.
				if (DataFile(filename, p_HW[devcnt]->DMA[chancnt-1].BurstSizeB * p_HW[devcnt]->DMA[chancnt-1].BurstCount, fsdivisor, 16) < 0)
					printf("\t[INFO] Discontinuities will be present in the channel #%u TX waveform.\n", chancnt);
				// Load DAC samples into the TX DMA buffer so that data is immediately available when the channel is started.
				if (DMALoad(filename, &p_HW[devcnt]->DMA[chancnt-1], 16))
					printf("\t[WARNING] Could not load DMA buffer contents from file %s.\n", filename);
				// Load the TX data buffer with samples to refresh the DMA buffer as data is consumed, the contents remain static in this example.
				if (DataLoad(filename, User[devcnt].p_DataBuffer[chancnt-1], p_HW[devcnt]->DMA[chancnt-1].BurstSizeB * p_HW[devcnt]->DMA[chancnt-1].BurstCount, p_HW[devcnt]->DMA[chancnt-1].PagesPerMark, 16))
					printf("\t[WARNING] Could not load TX data buffer contents from file %s.\n", filename);
			}
		}
	}

	///////////////////////////////////////////////  Hardware Performance Check  ///////////////////////////////////////////////
	
	// Measure hardware performance of each Red Rapids device and report the results.
	// Note: The application will still function correctly if this code segmented is removed.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
	{
		printf("\nChecking device DEV%u hardware performance...\n", devcnt);

		// Perform a limited memory test if the product includes attached QDR II+ SRAM.
		if (MemSize(p_HW[devcnt]))
		{
			for (index = 1; index < 3; index++)
				if ((status = MemTest(p_HW[devcnt],index)) < 0)
					printf("\t[ERROR] SRAM #%u memory test failed to complete.\n", index);
				else if (!status)
					printf("\t[INFO] SRAM #%u memory test passed.\n", index);
				else
					printf("\t[ERROR] SRAM #%u recorded %u memory errors\n", index, status);
		}
	
		// Measure the maximum achievable PCI throughput by initiating continuous DMA transactions on all channels.
		PciBenchmark(p_HW[devcnt]);
		// Retrieve the measured performance.
		PciPerformance(p_HW[devcnt], performance);
		// Report the measured instantaneous PCI throughput.
		if (User[devcnt].RXChannelCount) printf("\t[INFO] Measured RX PCI performance: %.3f Gbytes/sec\n", performance[0]);
		if (User[devcnt].TXChannelCount) printf("\t[INFO] Measured TX PCI performance: %.3f Gbytes/sec\n", performance[1]);
		// Report the measured maximum PCI latency recorded.
		if (User[devcnt].RXChannelCount) printf("\t[INFO] Measured RX PCI latency: %.3f usec\n", performance[2]);
		if (User[devcnt].TXChannelCount) printf("\t[INFO] Measured TX PCI latency: %.3f usec\n", performance[3]);
		// Clear the PCI latency performance measurement.
		ClearLatency(p_HW[devcnt]);
	}

	///////////////////////////////////////////////////  Process Channels  ///////////////////////////////////////////////////

	// Start processing data on all available channels of each Red Rapids device simultaneously.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
	{
		printf("\nStarting device DEV%u channels...\n", devcnt);
		
		// Enable each channel to start processing when the specified start event occurs.
		for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
		{
			// Clear the channel status register and error counters.
			ChannelStatusClear(p_HW[devcnt], chancnt);
			for (index = 0; index < 4; index++)
				User[devcnt].Error[chancnt-1][index] = 0;
			// Enable the channel to respond to the selected start event (software trigger).
			FlowControl(p_HW[devcnt], chancnt, 1);
			printf("\t[INFO] Channel #%u enabled.\n", chancnt);
		}

		// Monitor the status of all channels with a polling loop to determine when they are ready for a start event.
		do
		{
			status = 0;
			// Pause before the next status query to minimize PCI bus traffic.
			MicrosecondTimer(100);
			// Check the ready status of each channel (0 = Not Ready, 1 = Ready).
			for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
				status += ReadyStatus(p_HW[devcnt], chancnt);
		}
		while (status != channel_count[devcnt]);
	}

	// Display an update message that includes the interrupt status.
	if (GlobalMaskDefault)
		printf("\nWaiting for all channels to stop processing (interrupts enabled)...\n");
	else
		printf("\nWaiting for all channels to stop processing (interrupts disabled)...\n");
	// It is not safe to abort execution while DMA transactions are active.
	printf("\t[WARNING] CTRL-C FROM THE KEYBOARD MAY CRASH THE COMPUTER!\n");

	// Issue a software trigger to each Red Rapids device to start processing data on all channels.
	// Note: Different start/stop triggers can be selected in the channel configuration.
	// Note: Begin monitoring channel status immediately following the trigger so that DMA data is not lost.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
		SoftwareTrigger(p_HW[devcnt]);

	// If global interrupts are enabled, process channels based on an interrupt service routine (ISR).
	if (GlobalMaskDefault)
		// Run continuosly until all channels have been disabled by the selected stop event (sample count).
		do
		{
			status = 0;
			// It is customary to run each ISR in a separate thread, but that is not done in this demonstration for simplicity.
			for (devcnt = 0; devcnt < devsigfpga; devcnt++)
			{
				// Check the on/off status of each channel to determine whether any are still active.
				for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
					status += FlowStatus(p_HW[devcnt], chancnt);
				// Wait for the next interrupt if any channel is active.
				// Note: The select() function can be used instead of the read() function to avoid blocking.
				if (status && read(p_HW[devcnt]->DEV.fd, NULL, 4))
					{
						// Transfer data between the DMA buffer and the user defined data buffer.
						// Note: This function is not needed in applications that process data directly from the DMA buffer.
						ProcessChannels(p_HW[devcnt]);
						// Re-enable global interrupts at the end of the ISR.
						// Note: The UIO driver disables further interrupts until the current interrupt is serviced.
						InterruptMask(p_HW[devcnt], 1);
					}
			}
		} while (status);
	// If global interrupts are disabled, process channels based on a status polling loop.
	else
		// Run continuosly until all channels have been disabled by the selected stop event (sample count).
		do
		{
			status = 0;
			// It is customary to run each loop in a separate thread, but that is not done in this demonstration for simplicity.
			for (devcnt = 0; devcnt < devsigfpga; devcnt++)
			{
				// Check the on/off status of each channel to determine whether any are still active.
				for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
					status += FlowStatus(p_HW[devcnt], chancnt);
				// Transfer any new data between the DMA buffer and the user defined data buffer.
				// Note: This function is not needed in applications that process data directly from the DMA buffer.
				ProcessChannels(p_HW[devcnt]);
			}
			// Pause before the next channel status query to minimize PCI bus traffic.
			// Note: The pause should be much less than the time interval between DMA markers.
			if (status) MicrosecondTimer(100);
		} while (status);

	/////////////////////////////////////////////////////  Report Status  /////////////////////////////////////////////////////

	// Report hardware status retrieved from each Red Rapids device.
	// Note: The application will still function correctly if this code segmented is removed.
	for (devcnt = 0; devcnt < devsigfpga; devcnt++)
	{
		printf("\nRetrieving device DEV%u temperature status...\n", devcnt);
		// Report the values measured by the temperature monitor.
		TemperatureStatus(p_HW[devcnt], &Temperature);
		printf("\t[INFO] FPGA die temperature is %.2f degrees Celsius.\n", Temperature.Sensor[0]);
		printf("\t[INFO] ADC proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[1]);
		printf("\t[INFO] QDR proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[2]);
		printf("\t[INFO] Shielded proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[3]);
		printf("\t[INFO] Card center proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[4]);

		printf("\nRetrieving device DEV%u power status...\n", devcnt);
		// Report the values measured by the power monitor.
		printf("\t[INFO] Total power dissipation is %.2f Watts.\n", PowerStatus(p_HW[devcnt], &Power));
		printf("\t[INFO] Supply #1 summary: %.2f Volts, %.2f Amps, %.2f Watts\n", Power.Supply1[0], Power.Supply1[1], Power.Supply1[2]);
		printf("\t[INFO] Supply #2 summary: %.2f Volts, %.2f Amps, %.2f Watts\n", Power.Supply2[0], Power.Supply2[1], Power.Supply2[2]);
		// The third power supply is not present on native PCI Express products, do not report if the measured current is zero.
		if (Power.Supply3[1]) printf("\t[INFO] Supply #3 summary: %.2f Volts, %.2f Amps, %.2f Watts\n", Power.Supply3[0], Power.Supply3[1], Power.Supply3[2]);

		printf("\nReviewing device DEV%u channel status...\n", devcnt);
		// Report any faults that were detected during execution.
		ErrorStatus(p_HW[devcnt], &Error);
		if (Error.Bar[0]) printf("\t[ERROR] Write to an illegal BAR0 address detected.\n");
		if (Error.Bar[1]) printf("\t[ERROR] Read from an illegal BAR0 address detected.\n");
		if (Error.Bar[2]) printf("\t[ERROR] Write to an illegal BAR2 address detected.\n");
		if (Error.Bar[3]) printf("\t[ERROR] Read from an illegal BAR2 address detected.\n");
		if (Error.Clock[0]) printf("\t[ERROR] 200 MHz reference out of lock.\n");
		if (Error.Clock[1]) printf("\t[ERROR] Flash programmer clock out of lock.\n");
		if (Error.Clock[2]) printf("\t[ERROR] QDR-A SRAM clock out of lock.\n");
		if (Error.Clock[3]) printf("\t[ERROR] QDR-B SRAM clock out of lock.\n");
		if (Error.Clock[4] || Error.Clock[5] || Error.Clock[6] || Error.Clock[7]) printf("\t[ERROR] RX clock out of lock.\n");

		// Report any channel errors that were detected during execution.
		for (chancnt = 1; chancnt <= channel_count[devcnt]; chancnt++)
		{
			if (User[devcnt].Error[chancnt-1][0]) printf("\t[ERROR] Channel #%u converter (ADC/DAC) error detected.\n", chancnt);
			if (User[devcnt].Error[chancnt-1][1]) printf("\t[ERROR] Channel #%u FIFO overflow/underflow, check PCI performance.\n", chancnt);
			if (User[devcnt].Error[chancnt-1][2]) printf("\t[ERROR] Channel #%u DMA data was lost, adjust the DMA marker to correct.\n", chancnt);
		}

		// Report the interrupt response time if interrupts were enabled.
		if (GlobalMaskDefault) printf("\t[INFO] The last interrupt was serviced in %.2f microseconds.\n", InterruptResponse(p_HW[devcnt]));

		// Extract the contents of each RX buffer and save to a file.
		if (User[devcnt].RXChannelCount)
		{
			printf("\nSaving captured data on all device DEV%u receiver channels...\n", devcnt);
			for (chancnt = 1; chancnt <= User[devcnt].RXChannelCount; chancnt++)
			{
				// Save the first page of each DMA buffer to a file.
				sprintf(filename, "dmabuf%u-%u.txt", devcnt, chancnt);
				printf("\t[INFO] Writing channel #%u DMA buffer contents to file %s.\n", chancnt, filename);
				if (DMASave(filename, &p_HW[devcnt]->DMA[chancnt-1], 1, 16))
					printf("\t[WARNING] Could not save DMA buffer contents to file %s.\n", filename);
				// Save the first page of each RX data buffer to a file.
				sprintf(filename, "rxdata%u-%u.txt", devcnt, chancnt);
				printf("\t[INFO] Writing channel #%u data buffer contents to file %s.\n", chancnt, filename);
				if (DataSave(filename, User[devcnt].p_DataBuffer[chancnt-1], p_HW[devcnt]->DMA[chancnt-1].BurstSizeB * p_HW[devcnt]->DMA[chancnt-1].BurstCount, 1, 16))
					printf("\t[WARNING] Could not save RX data buffer contents to file %s.\n", filename);
			}
		}
	}
		
	////////////////////////////////////////////////  Free Allocated Resources ////////////////////////////////////////////////

	// Free all of the host resources allocated to each Red Rapids device.
	for (index = 0; index < devsigfpga; index++)
	{
		printf("\nFreeing resources allocated to device DEV%u...\n", index);
		status += FreeResources(p_HW[index], channel_count[index], UIO[index]);
	}
	if (status)
		printf("\t[ERROR] Some allocated resources could not be freed.\n");

	printf("\n...all operations completed.\n");
	return EXIT_SUCCESS;
}
