/****************************************************  Channel Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are specific to individual SigFPGA channels.

	Rev History:  03-11-2019:	Some function and variable names changed for consistency across products.
				  11-28-2018:	Style and format changes.
				  08-08-2018:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Channel Functions  *****************************************************/

#include "driver_functions.h"
#include "sigfpga.h"
#include "channel_functions.h"

// Set the channel interrupt mask bits to enable (1) or disable (0) interrupt generation for each event type.
void ChannelMask(s_Handle *p_HW, unsigned ChannelNumber, s_ChannelIntMask ChannelIntMask)
{
	// Write to the channel interrupt mask address.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + CHANNEL_INT_MASK] =
				(unsigned long long)ChannelIntMask.DMAMarker |
				(unsigned long long)ChannelIntMask.Error[0] << 4 |
				(unsigned long long)ChannelIntMask.Error[1] << 5;
}

// Report the channel status and clear the values.
// Returns the total number of status events detected.
int ChannelStatus(s_Handle *p_HW, unsigned ChannelNumber, s_ChannelStatus *p_Channel)
{
	unsigned long long value = 0;
	int status = 0;
	unsigned cnt = 0;
	
	// Save and clear the contents of the channel status register by writing to the error status address.
	// Note: The contents are saved to a shadow register that is accessed during a read operation (below).
	p_HW->DEV.BAR2[(512 * ChannelNumber) + CHANNEL_STATUS] = 0x0ull;
	// Read the saved contents of the channel status.
	value = p_HW->DEV.BAR2[(512 * ChannelNumber) + CHANNEL_STATUS];

	// Strip the DMA marker bit.
	if ((p_Channel->DMAMarker = value & 0x0000000000000001ull)) status++;

	// Strip the individual channel error indicator bits.
	for (cnt = 0; cnt < 2; cnt++)
		if ((p_Channel->Error[cnt] = value >> ((cnt + 4)) & 0x0000000000000001ull)) status++;

	// Return the total fault count.
	return status;
}

// Clear the channel status register.
void ChannelStatusClear(s_Handle *p_HW, unsigned ChannelNumber)
{
	// Write to the channel status address.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + CHANNEL_STATUS] = 0x0ull;
}

// Enable (OnOff = 1) or disable (OnOff = 0) channel processing.
void FlowControl(s_Handle *p_HW, unsigned ChannelNumber, unsigned OnOff)
{
	// Write to the synchronizer control address.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + SYNCHRONIZER_CTL] = (unsigned long long)OnOff;
}

// Report the current on/off status of the requested channel.
// Return of zero indicates the channel is off, one indicates the channel is on.
int FlowStatus(s_Handle *p_HW, unsigned ChannelNumber)
{
	// Read the synchronizer control address.
	return (int)(p_HW->DEV.BAR2[(512 * ChannelNumber) + SYNCHRONIZER_CTL] & 0x0000000000000001ull);
}

// Report the current ready status of the requested channel.
// Return of zero indicates the channel is not ready for a start trigger, one indicates the channel is ready.
int ReadyStatus(s_Handle *p_HW, unsigned ChannelNumber)
{
	// Read the synchronizer control address.
	// Note: Shift the ready status bit so that the function returns either one or zero.
	return (int)((p_HW->DEV.BAR2[(512 * ChannelNumber) + SYNCHRONIZER_CTL] >> 1) & 0x0000000000000001ull);
}

// Configure the synchronizer operating controls.
void SynchronizerSettings(s_Handle *p_HW, unsigned ChannelNumber, s_Synchronizer Synchronizer)
{
	// Write to the synchronizer command address.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + SYNCHRONIZER_CMD] =
			(unsigned long long)Synchronizer.StartEvent |
			(unsigned long long)Synchronizer.StopEvent << 2 |
			(unsigned long long)Synchronizer.SampleCount << 32;
}
