/*****************************************************  Driver Functions  ******************************************************
	Red Rapids
	797 N. Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	Description: Adapter driver function definitions.

	Rev History: 01-20-2020:	Modified UIOScan() to return the total number Red Rapids devices detected.
	Rev History: 12-09-2019:	Increased the size of pathname[] variable in function DMAOpen().
				 11-28-2018:	Moved DMA page number counter from DMAMap() to DMAUnmap(), fixed incorrect BAR reference in UIOScan().
				 08-08-2018:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*****************************************************  Driver Functions  ******************************************************/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "driver_functions.h"

// Unmap the UIO character device BAR address spaces.
// Return of zero indicates success, non-zero indicates an error condition.
int DEVClose(s_UIOScan UIO, s_DEVMap *p_DEV)
{
	// Delete the BAR0 mapping. 
	if (p_DEV->BAR0 != NULL && p_DEV->BAR0 != MAP_FAILED)
		if (munmap((void *)p_DEV->BAR0, UIO.BAR0Size) < 0)
			return errno;

	// Delete the BAR2 mapping. 
	if (p_DEV->BAR2 != NULL && p_DEV->BAR2 != MAP_FAILED)
		if (munmap((void *)p_DEV->BAR2, UIO.BAR2Size) < 0)
			return errno;
		
	// Close the character file. 
	if (close(p_DEV->fd) < 0)
		return errno;

	return 0;
}

// Initializes all varibles in the device handle structure to default values.
void DEVInit(s_Handle *p_HW, unsigned DevNum)
{
	// Set a unique device number for each hardware handle.
	p_HW->DevNum = DevNum;
	// Set the file descriptor to an illegal value.
	p_HW->DEV.fd = -1;
	// Set all pointers to NULL.
	p_HW->DEV.BAR0 = NULL;
	p_HW->DEV.BAR2 = NULL;
	p_HW->DMA = NULL;
	p_HW->User = NULL;
}
	
// Open the UIO character device file and memory map the BAR address spaces to virtual addresses in user space.
// Return of zero indicates success, non-zero indicates an error condition.
int DEVOpen(s_UIOScan UIO, s_DEVMap *p_DEV)
{
	char uio_path[11];
		
	// Create a string that includes the full directory path name to the requested UIO device character file.
	sprintf(uio_path, "/dev/uio%u", UIO.UIONumber);
					
	// Open the requested UIO device character file.
	if ((p_DEV->fd = open(uio_path, O_RDWR|O_SYNC)) < 0)
		return errno;
	
	// Memory map the BAR0 address space.
	if (UIO.BAR0Size)
		if ((p_DEV->BAR0 = (unsigned long long *)mmap(NULL, UIO.BAR0Size, PROT_READ|PROT_WRITE, MAP_SHARED, p_DEV->fd, 0)) == MAP_FAILED)
			return errno;

	// Memory map the BAR2 address space.
	if (UIO.BAR2Size)
		if ((p_DEV->BAR2 = (unsigned long long *)mmap(NULL, UIO.BAR2Size, PROT_READ|PROT_WRITE, MAP_SHARED, p_DEV->fd, sysconf(_SC_PAGESIZE))) == MAP_FAILED)
			return errno;

	return 0;
}

// Unmap the UIO character device BAR address spaces.
// Return of zero indicates success, non-zero indicates an error condition.
int DMAClose(int fd)
{
	// Close the character file. 
	if (close(fd) < 0)
		return errno;
	
	return 0;
}

// Set the DMA driver page counter to open page zero on the next mapping.
// Return of zero indicates success, non-zero indicates an error condition.
int DMAFirstPage(s_DMABuf *p_DMA)
{
	void *p_address;
	
	while(p_DMA->PageNumber != p_DMA->PageCount - 1)
	{
		// Increment the page number.
		p_DMA->PageNumber++;
		// Memory map the next available page.
		// Note: Only request a single Linux page during this process since it will not be used.
		if ((p_address = mmap(0, sysconf(_SC_PAGESIZE), PROT_READ|PROT_WRITE, MAP_SHARED|MAP_LOCKED, p_DMA->fd, 0)) == MAP_FAILED)
			return errno;
		// Unmap the page.
		if (munmap(p_address, sysconf(_SC_PAGESIZE)) < 0)
			return errno;
	}
	
	return 0;
}

// Initializes all varibles in the s_DMABuf structure to default values.
void DMAInit(s_DMABuf *p_DMA)
{
	// Set the file descriptor to an illegal value.
	p_DMA->fd = -1;
	// Set all of the size and count variables to zero.
	p_DMA->PageNumber = 0;
	p_DMA->PageCount = 0;
	p_DMA->PagesPerMark = 0;
	p_DMA->BurstSizeB = 0;
	p_DMA->BurstCount = 0;
	// Set all pointers to NULL.
	p_DMA->Address = NULL;
}

// Retrieve the virtual address of the next available DMA page.
// Return of zero indicates success, non-zero indicates an error condition.
int DMAMap(s_DMABuf *p_DMA)
{
	// Memory map the next page to be serviced.
	if ((p_DMA->Address =(unsigned long long *)mmap(0, p_DMA->BurstSizeB * p_DMA->BurstCount, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_LOCKED, p_DMA->fd, 0)) == MAP_FAILED)
		return errno;
	
	return 0;
}

// Open the DMA character device file and retrieve the buffer configuration information.
// Return of zero indicates success, non-zero indicates an error condition.
int DMAOpen(s_DMABuf *p_DMA, unsigned DeviceNumber, unsigned BufferIndex)
{
	char pathname[16];
	void *p_address;
	unsigned long long *p_dmapage;
	
	// Create a string that includes the full directory path name to the requested DMA device character file.
	sprintf(pathname, "/dev/rrdma%u-%u", DeviceNumber, BufferIndex);
	
	// Open the requested DMA device character file.
	// Note: This action will set the next page to zero.
	if ((p_DMA->fd = open(pathname, O_RDWR|O_SYNC)) < 0)
		return errno;
	
	// Query the DMA buffer to retrieve configuration information.
	do
	{
		// Memory map the next available page.
		// Note: Only request a single Linux page during this process since the actual size is unknown.
		if ((p_address = mmap(0, sysconf(_SC_PAGESIZE), PROT_READ|PROT_WRITE, MAP_SHARED|MAP_LOCKED, p_DMA->fd, 0)) == MAP_FAILED)
			return errno;
		p_dmapage = (unsigned long long *)p_address;
		p_DMA->PageNumber = (unsigned)p_dmapage[0];
		// Record the variables that are constant across all pages when the first page is mapped.
		if (p_DMA->PageNumber == 0)
		{
			p_DMA->PageCount = (unsigned)p_dmapage[1];
			p_DMA->PagesPerMark = (unsigned)p_dmapage[2];
			p_DMA->BurstSizeB = (unsigned)p_dmapage[3];
			p_DMA->BurstCount = (unsigned)p_dmapage[4];
			// Allocate memory to record the physical address of each page in the buffer.
			if ((p_DMA->Address = malloc(sizeof *p_DMA->Address * p_DMA->PageCount)) == NULL)
				return errno;
		}
		// Record the physical address of every page.
		p_DMA->Address[p_DMA->PageNumber] = p_dmapage[5];
		// Unmap the page.
		if (munmap(p_address, sysconf(_SC_PAGESIZE)) < 0)
			return errno;
	} while(p_DMA->PageNumber < p_DMA->PageCount - 1);
	
	return 0;
}

// Unmap the UIO character device BAR address spaces.
// Return of zero indicates success, non-zero indicates an error condition.
int DMAUnmap(s_DMABuf *p_DMA)
{
	// Increment the page number.
	if (p_DMA->PageNumber < p_DMA->PageCount - 1)
		p_DMA->PageNumber++;
	else
		p_DMA->PageNumber = 0;
	
	// Unmap the virtual address of the current DMA page.
	if (munmap(p_DMA->Address, p_DMA->BurstSizeB * p_DMA->BurstCount) < 0)
		return errno;
			
	return 0;
}

// Scan all available UIO devices for any Red Rapids device.
// Returns the total number of Red Rapids devices detected.
unsigned UIOScan(s_UIOScan *p_UIO)
{
	FILE *fp;
	unsigned devcnt = 0;
	unsigned uiocnt = 0;
	char uio_path[21];
	char file_name[45];
	char inputline[6];
	
	// Initialize each member of the hardware descriptor structure.
	for (devcnt = 0; devcnt < NUM_DEVICES; devcnt++)
	{
		p_UIO[devcnt].UIONumber = 0;
		p_UIO[devcnt].SubSysID = 0;
		p_UIO[devcnt].BAR0Size = 0;
		p_UIO[devcnt].BAR2Size = 0;
	}
	
	devcnt = 0;
	
	// Scan all of the available UIO devices.
	do
	{
		// Create a string that includes the full path to the next possible UIO directory.
		sprintf(uio_path, "/sys/class/uio/uio%u", uiocnt);
		
		// Create a string that includes the full path to the "name" file.
		strcpy(file_name, uio_path);
		strcat(file_name, "/name");
		
		// Attempt to open the next UIO device "name" file.
		fp = fopen(file_name, "r");

		// Extract hardware information from the UIO device file system.
		if (fp != NULL)
			if (fgets(inputline, 6, fp) != NULL)
			{
				// Confirm that this is a Red Rapids device.
				if (!strcmp(inputline, "rrdev"))
				{
					fclose(fp);
					// Record the UIO device number.
					p_UIO[devcnt].UIONumber = devcnt;
					
					// Record the subsystem device ID.
					strcpy(file_name, uio_path);
					strcat(file_name, "/device/subsystem_device");
					if ((fp = fopen(file_name, "r")) == NULL)
						p_UIO[devcnt].SubSysID = 0;
					else
					{
						fscanf(fp, "%x", &p_UIO[devcnt].SubSysID);
						fclose(fp);
					}
					
					// Record the BAR0 size.
					strcpy(file_name, uio_path);
					strcat(file_name, "/maps/map0/size");
					if ((fp = fopen(file_name, "r")) == NULL)
						p_UIO[devcnt].BAR0Size = 0;
					else
					{
						fscanf(fp, "%llx", &p_UIO[devcnt].BAR0Size);
						fclose(fp);
					}
					
					// Record the BAR2 size.
					strcpy(file_name, uio_path);
					strcat(file_name, "/maps/map1/size");
					if ((fp = fopen(file_name, "r")) == NULL)
						p_UIO[devcnt].BAR2Size = 0;
					else
					{
						fscanf(fp, "%llx", &p_UIO[devcnt].BAR2Size);
						fclose(fp);
					}
					
					// Increment the device count.
					devcnt++;
				}
			else
				fclose(fp);
			}

		// Increment the UIO count.
		uiocnt++;
		
	} while (fp != NULL);
	
	return devcnt;
}
