/****************************************************  Product Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to initialize the hardware specific to each SigFPGA model.

	Rev History:  02-14-2020:	Added a write operation to the AD9142A DAC PLL enable sequence to eliminate lock problems.
				  12-09-2019:	Moved SoftwareReset() to MxxxInitialize() since it is synchronous and relies on a stable clock.
				  03-11-2019:	Some function and variable names changed for consistency across products.
				  11-28-2018:	Style and format changes.
				  08-08-2018:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Product Functions  *****************************************************/

#include <stdlib.h>

#include "driver_functions.h"
#include "sigfpga.h"
#include "global_functions.h"
#include "chip_functions.h"
#include "product_functions.h"
#include "default_hardware_config.h"

// Intitialize the Red Rapids hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int HWInitialize(s_Handle *p_HW, unsigned Model)
{
	switch (Model)
	{
		case 0x370:	// Model 370
			// Initialize hardware specific to the Model 376.
			return M370Initialize(p_HW, M370Default);
		case 0x371:	// Model 371
			// Initialize hardware specific to the Model 371.
			return M371Initialize(p_HW, M371Default);
		case 0x372:	// Model 372
			// Initialize hardware specific to the Model 372.
			return M372Initialize(p_HW, M372Default);
		case 0x373:	// Model 373
			// Initialize hardware specific to the Model 373.
			return M373Initialize(p_HW, M373Default);
		case 0x376:	// Model 376
			// Initialize hardware specific to the Model 376.
			return M376Initialize(p_HW, M376Default);
		case 0x377:	// Model 377
			// Initialize hardware specific to the Model 377.
			return M377Initialize(p_HW, M377Default);
		case 0x378:	// Model 378
			// Initialize hardware specific to the Model 378.
			return M378Initialize(p_HW, M378Default);
		default:	// Illegal model number.
			return MODEL_NUMBER;
	}
}

// Intitialize the Model 370 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M370Initialize(s_Handle *p_HW, s_M370Devices M370Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The standard Model 370 has one receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The SigFPGA hardware sign extends the 12-bit sample data to a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the ADC from the default structure values.
	for (cnt = 0; cnt < 11; cnt++)
		ADC12D1600Write(p_HW, M370_RX1_ADC, M370Devices.RX1_ADC.Payload[cnt][0], M370Devices.RX1_ADC.Payload[cnt][1]);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Configure the trim DAC from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
		LTC1661Write(p_HW, M370_RX1_TRIM_DAC, M370Devices.RX1_TRIM_DAC.Payload[cnt][0], M370Devices.RX1_TRIM_DAC.Payload[cnt][1]);

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		ADL5566Write(p_HW, CMD_STAT1, M370Devices.RX1_AMP_ENBL.Enable[cnt][0], M370Devices.RX1_AMP_ENBL.Enable[cnt][1]);
	
	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}

// Intitialize the Model 371 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M371Initialize(s_Handle *p_HW, s_M371Devices M371Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The Model 371 is equipped with two receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 371 RX channels produce a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the dual ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		ADS42LB69Write(p_HW, M371_RX1TO2_ADC, M371Devices.RX1TO2_ADC.Payload[cnt][0], M371Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M371_RX1_TRIM_DAC, M371Devices.RX1_TRIM_DAC.Payload[cnt][0], M371Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M371_RX2_TRIM_DAC, M371Devices.RX2_TRIM_DAC.Payload[cnt][0], M371Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		ADL5566Write(p_HW, CMD_STAT1, M371Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M371Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);

	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the selected input to the clock distribution chip.
		// Note: The default input is the on-board synthesizer.
		M371Devices.ClockDist.Payload[9][1] = 0x02;

	// Hardware reset the clock distribution chip.
	AD9512Write(p_HW, M371_CLK_DIST, 0x0000, 0x30);
	AD9512Write(p_HW, M371_CLK_DIST, 0x005A, 0x01);
	AD9512Write(p_HW, M371_CLK_DIST, 0x0000, 0x10);
	AD9512Write(p_HW, M371_CLK_DIST, 0x005A, 0x01);
	// Configure the clock distribution chip by loading the default values.
	for (cnt = 0; cnt < 22; cnt++)
		AD9512Write(p_HW, M371_CLK_DIST, M371Devices.ClockDist.Payload[cnt][0], M371Devices.ClockDist.Payload[cnt][1]);
	
	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}

// Intitialize the Model 372 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M372Initialize(s_Handle *p_HW, s_M372Devices M372Devices)
{
	unsigned status = 0;
	unsigned tx_refclk = 0;
	unsigned tx_dacclk = 0;
	unsigned loop_divider = 0;
	unsigned vco_divider = 1;
	unsigned diglogic_divider = 0;
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Query the hardware for the number of receiver and transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 372 RX and TX channels process a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 16;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Reset the clock distribution chip.
	LMK01000Write(p_HW, M372_CLK_DIST, 0x0, 0x8000010);
	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the clock distribution chip input to CLKin1.
		// Note: The default input is the on-board synthesizer (CLKin0).
		M372Devices.ClockDist.Payload[9][1] &= 0xDFFFFFF;
	// Configure the clock distribution chip by loading the modified default values.
	for (cnt = 0; cnt < 10; cnt++)
		LMK01000Write(p_HW, M372_CLK_DIST, M372Devices.ClockDist.Payload[cnt][0], M372Devices.ClockDist.Payload[cnt][1]);
	// Synchronize the outputs of the clock distribution chip.
	LMK01000Sync(p_HW, M372_CLK_DIST_SYNC);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Reset the dual ADC that is assigned to RX channels #1 and #2.
	AD9652Write(p_HW, M372_RX1TO2_ADC, 0x0, 0x3C);
	// Wait 500 us for the SPI interface to become available after reset.
	MicrosecondTimer(500);
	// Configure the dual ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 26; cnt++)
		AD9652Write(p_HW, M372_RX1TO2_ADC, M372Devices.RX1TO2_ADC.Payload[cnt][0], M372Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the RX trim DACs from the default structure values.
	// Note: Trim DACs are only present on RX DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M372_RX1_TRIM_DAC, M372Devices.RX1_TRIM_DAC.Payload[cnt][0], M372Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M372_RX2_TRIM_DAC, M372Devices.RX2_TRIM_DAC.Payload[cnt][0], M372Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the RX amplifier enables from the default structure values.
	// Note: Amplifiers are only present on RX DC coupled units.
	ADA4927Write(p_HW, CMD_STAT1, M372Devices.RX1_AMP_ENBL.Enable[0], M372Devices.RX1_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M372Devices.RX2_AMP_ENBL.Enable[0], M372Devices.RX2_AMP_ENBL.Enable[1]);

	// Phase align the clocks to valid data.
	if ((status = ClockInitLMK01000(p_HW, M372Default.ClockDist.Payload[0][1], M372Default.ClockDist.Payload[4][1], M372Default.ClockDist.Payload[5][1])))
		return status;
	
	// Initiate the ADC fast start-up calibration.
	AD9652Cal(p_HW, M372_RX1TO2_ADC);
	
	// Measure the DAC REFCLK frequency to compute DLL and PLL settings.
	tx_refclk = ClockFrequency(p_HW);

	// Compute the maximum DAC interpolation setting allowed with the measured REFCLK frequency if requested.
	if (M372Devices.TX3TO4_DAC.Payload[27][1] == 0x20)
	{
		// Select the maximum interpolation rate allowed at the DCI (REFCLK) frequency and below the PLL maximum operating frequency (2.07 GHz).
		if ((tx_refclk <= 575000000) && (2 * tx_refclk <= 2070000000))
		{
			M372Devices.TX3TO4_DAC.Payload[27][1] = 0;
			tx_dacclk = 2 * tx_refclk;
			loop_divider = 0;
		}
		if ((tx_refclk <= 375000000) && (4 * tx_refclk <= 2070000000))
		{
			M372Devices.TX3TO4_DAC.Payload[27][1] = 2;
			tx_dacclk = 4 * tx_refclk;
			loop_divider = 1;
		}
		if ((tx_refclk <= 187500000) && (8 * tx_refclk <= 2070000000))
		{
			M372Devices.TX3TO4_DAC.Payload[27][1] = 3;
			tx_dacclk = 8 * tx_refclk;
			loop_divider = 2;
		}
		if (M372Devices.TX3TO4_DAC.Payload[27][1] == 0x20)
			return DAC_INTERPOLATE;
	}
	// Use the specified DAC interpolation setting.
	else
		switch (M372Devices.TX3TO4_DAC.Payload[27][1])
		{
			case 0: tx_dacclk = 2 * tx_refclk; loop_divider = 0; break;
			case 2: tx_dacclk = 4 * tx_refclk; loop_divider = 1; break;
			case 3: tx_dacclk = 8 * tx_refclk; loop_divider = 2; break;
			default: return DAC_INTERPOLATE;
		}

	// Compute the DAC PLL settings if requested and the PLL is enabled.
	if ((M372Devices.TX3TO4_DAC.Payload[14][1] == 0x20) && (M372Devices.TX3TO4_DAC.Payload[12][1] == 0x80))
	{
		// Set the highest VCO frequency within the optimal operating range (1.03 GHz to 2.07 GHz).
		if ((tx_dacclk >= 1030000000) && (tx_dacclk <= 2070000000)) vco_divider = 0;
		if ((2 * tx_dacclk >= 1030000000) && (2 * tx_dacclk <= 2070000000)) vco_divider = 4;
		if ((4 * tx_dacclk >= 1030000000) && (4 * tx_dacclk <= 2070000000)) vco_divider = 8;
		// Report an error if there is no solution possible at the measured DACCLK frequency.
		if (vco_divider == 1) return DAC_PLL;
		// Set the highest PLL digital clock possible (< 75 MHz).
		if (tx_refclk / 16 < 75000000) diglogic_divider = 192;
		if (tx_refclk / 8 < 75000000) diglogic_divider = 128;
		if (tx_refclk / 4 < 75000000) diglogic_divider = 64;
		if (tx_refclk / 2 < 75000000) diglogic_divider = 0;
		// Create the composite PLL setting.
		M372Devices.TX3TO4_DAC.Payload[14][1] = diglogic_divider + vco_divider + loop_divider;
	}

	// Select DLL or Delay Line interface mode if requested.
	if (M372Devices.TX3TO4_DAC.Payload[7][1] == 0x20)
	{
		// The DLL requres a minimum 250 MHz DCI (REFCLK) frequency.
		if (tx_refclk >= 250000000)
			M372Devices.TX3TO4_DAC.Payload[7][1] = 0xC0;
		else
			M372Devices.TX3TO4_DAC.Payload[7][1] = 0x00;
	}

	// Configure the data control settings based on DCI (REFCLK) frequency if requested.
	if (M372Devices.TX3TO4_DAC.Payload[9][1] == 0x26)
	{
		// DLL interface mode is selected.
		if (M372Devices.TX3TO4_DAC.Payload[7][1] >= 128)
			// Determine whether the DCI (REFCLK) is 350 MHz or higher.
			if (tx_refclk >= 350000000)
				M372Devices.TX3TO4_DAC.Payload[9][1] = 0x06;
			else
				M372Devices.TX3TO4_DAC.Payload[9][1] = 0x86;
		// Delay Line interface mode is selected.
		else
			M372Devices.TX3TO4_DAC.Payload[9][1] = 0x16;
	}

	// Configure the DLL delay settings (two registers) if requested.
	if ((M372Devices.TX3TO4_DAC.Payload[51][1] == 0x20) || (M372Devices.TX3TO4_DAC.Payload[52][1] == 0x20))
	{
		// DLL interface mode is selected.
		if (M372Devices.TX3TO4_DAC.Payload[7][1] >= 128)
		{
			M372Devices.TX3TO4_DAC.Payload[51][1] = 0xFE;
			M372Devices.TX3TO4_DAC.Payload[52][1] = 0x67;
		}
		// Delay Line interface mode is selected.
		else
		{
			if (tx_refclk > 170000000)
			{
				M372Devices.TX3TO4_DAC.Payload[51][1] = 0x00;
				M372Devices.TX3TO4_DAC.Payload[52][1] = 0x60;
			}
			else if (tx_refclk > 150000000)
			{
				M372Devices.TX3TO4_DAC.Payload[51][1] = 0x80;
				M372Devices.TX3TO4_DAC.Payload[52][1] = 0x67;
			}
			else if (tx_refclk > 130000000)
			{
				M372Devices.TX3TO4_DAC.Payload[51][1] = 0xF0;
				M372Devices.TX3TO4_DAC.Payload[52][1] = 0x67;
			}
			else
			{
				M372Devices.TX3TO4_DAC.Payload[51][1] = 0xFE;
				M372Devices.TX3TO4_DAC.Payload[52][1] = 0x67;
			}
		}
	}

	// Reset the dual DAC that is assigned to TX channels #3 and #4.
	AD9142AWrite(p_HW, M372_TX3TO4_DAC, 0x00, 0x20);
	// Configure the dual DAC that is assigned to TX channels #3 and #4 by loading the default values.
	for (cnt = 0; cnt < 63; cnt++)
		AD9142AWrite(p_HW, M372_TX3TO4_DAC, M372Devices.TX3TO4_DAC.Payload[cnt][0], M372Devices.TX3TO4_DAC.Payload[cnt][1]);

	// Enable the PLL to operate in automatic mode if it is enabled.
	if (M372Devices.TX3TO4_DAC.Payload[12][1] == 0x80)
	{
		// Enable the DAC PLL automatic VCO band select feature.
		AD9142AWrite(p_HW, M372_TX3TO4_DAC, 0x12, 0x40);
		AD9142AWrite(p_HW, M372_TX3TO4_DAC, 0x12, 0xC0);
		AD9142AWrite(p_HW, M372_TX3TO4_DAC, 0x12, 0x80);
		// Wait 10 ms for the band select to complete.
		MicrosecondTimer(10000);
		// Check the DAC PLL lock status.
		if (!(AD9142ARead(p_HW, M372_TX3TO4_DAC, 0x16) & 0x80)) return DAC_PLL_LOCK;
	}

	if (M372Devices.TX3TO4_DAC.Payload[7][1] >= 128)
	// Check the DAC DLL lock status if DLL interface mode is selected.
	if (M372Devices.TX3TO4_DAC.Payload[7][1] >= 128)
		if (!(AD9142ARead(p_HW, M372_TX3TO4_DAC, 0x0E) & 0x80)) return DAC_DLL_LOCK;

	// Reset the DAC FIFO.
	AD9142AWrite(p_HW, M372_TX3TO4_DAC, 0x25, 0x01);
	// Check the DAC FIFO status.
	if (!(AD9142ARead(p_HW, M372_TX3TO4_DAC, 0x25) & 0x02)) return DAC_FIFO_RST;
	// Power up the DAC outputs.
	AD9142AWrite(p_HW, M372_TX3TO4_DAC, 0x01, 0x00);

	// Configure the TX trim DACs from the default structure values.
	// Note: Trim DACs are only present on TX DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M372_TX3_TRIM_DAC, M372Devices.TX3_TRIM_DAC.Payload[cnt][0], M372Devices.TX3_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M372_TX4_TRIM_DAC, M372Devices.TX4_TRIM_DAC.Payload[cnt][0], M372Devices.TX4_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the TX amplifier enables from the default structure values.
	// Note: Amplifiers are only present on TX DC coupled units.
	ADA4927Write(p_HW, CMD_STAT1, M372Devices.TX3_AMP_ENBL.Enable[0], M372Devices.TX3_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M372Devices.TX4_AMP_ENBL.Enable[0], M372Devices.TX4_AMP_ENBL.Enable[1]);

	return OK;
}

// Intitialize the Model 373 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M373Initialize(s_Handle *p_HW, s_M373Devices M373Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;
	int status = OK;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Query the hardware for the number of receiver and transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 373 RX channels process a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Reset the clock distribution chip.
	LMK01000Write(p_HW, M373_CLK_DIST, 0x0, 0x8000010);
	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the clock distribution chip input to CLKin1.
		// Note: The default input is the on-board synthesizer (CLKin0).
		M373Devices.ClockDist.Payload[9][1] &= 0xDFFFFFF;
	// Configure the clock distribution chip by loading the modified default values.
	for (cnt = 0; cnt < 10; cnt++)
		LMK01000Write(p_HW, M373_CLK_DIST, M373Devices.ClockDist.Payload[cnt][0], M373Devices.ClockDist.Payload[cnt][1]);
	// Synchronize the outputs of the clock distribution chip.
	LMK01000Sync(p_HW, M373_CLK_DIST_SYNC);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Reset the dual ADC that is assigned to RX channels #1 and #2.
	AD9652Write(p_HW, M373_RX1TO2_ADC, 0x0, 0x3C);
	// Wait 500 us for the SPI interface to become available after reset.
	MicrosecondTimer(500);
	// Configure the dual ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 26; cnt++)
		AD9652Write(p_HW, M373_RX1TO2_ADC, M373Devices.RX1TO2_ADC.Payload[cnt][0], M373Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the RX trim DACs from the default structure values.
	// Note: Trim DACs are only present on RX DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M373_RX1_TRIM_DAC, M373Devices.RX1_TRIM_DAC.Payload[cnt][0], M373Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M373_RX2_TRIM_DAC, M373Devices.RX2_TRIM_DAC.Payload[cnt][0], M373Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the RX amplifier enables from the default structure values.
	// Note: Amplifiers are only present on RX DC coupled units.
	ADA4927Write(p_HW, CMD_STAT1, M373Devices.RX1_AMP_ENBL.Enable[0], M373Devices.RX1_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M373Devices.RX2_AMP_ENBL.Enable[0], M373Devices.RX2_AMP_ENBL.Enable[1]);
	
	// Phase align the clocks to valid data.
	if ((status = ClockInitLMK01000(p_HW, M373Default.ClockDist.Payload[0][1], M373Default.ClockDist.Payload[4][1], M373Default.ClockDist.Payload[5][1])))
		return status;
	
	// Initiate the ADC fast start-up calibration.
	AD9652Cal(p_HW, M373_RX1TO2_ADC);
	
	return OK;
}

// Intitialize the Model 376 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M376Initialize(s_Handle *p_HW, s_M376Devices M376Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The standard Model 376 has two receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The SigFPGA hardware sign extends the 12-bit sample data to a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the ADC from the default structure values.
	for (cnt = 0; cnt < 11; cnt++)
		ADC12D1600Write(p_HW, M376_RX1TO2_ADC, M376Devices.RX1TO2_ADC.Payload[cnt][0], M376Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M376_RX1_TRIM_DAC, M376Devices.RX1_TRIM_DAC.Payload[cnt][0], M376Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M376_RX2_TRIM_DAC, M376Devices.RX2_TRIM_DAC.Payload[cnt][0], M376Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		ADL5566Write(p_HW, CMD_STAT1, M376Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M376Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);
	
	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}

// Intitialize the Model 377 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M377Initialize(s_Handle *p_HW, s_M377Devices M377Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The Model 377 can is equipped with four receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 377 RX channels produce a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the first ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		ADS42LB69Write(p_HW, M377_RX1TO2_ADC, M377Devices.RX1TO2_ADC.Payload[cnt][0], M377Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the second ADC that is assigned to RX channels #3 and #4 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		ADS42LB69Write(p_HW, M377_RX3TO4_ADC, M377Devices.RX3TO4_ADC.Payload[cnt][0], M377Devices.RX3TO4_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M377_RX1_TRIM_DAC, M377Devices.RX1_TRIM_DAC.Payload[cnt][0], M377Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M377_RX2_TRIM_DAC, M377Devices.RX2_TRIM_DAC.Payload[cnt][0], M377Devices.RX2_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M377_RX3_TRIM_DAC, M377Devices.RX3_TRIM_DAC.Payload[cnt][0], M377Devices.RX3_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M377_RX4_TRIM_DAC, M377Devices.RX4_TRIM_DAC.Payload[cnt][0], M377Devices.RX4_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
	{
		ADL5566Write(p_HW, CMD_STAT1, M377Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M377Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);
		ADL5566Write(p_HW, CMD_STAT1, M377Devices.RX3TO4_AMP_ENBL.Enable[cnt][0], M377Devices.RX3TO4_AMP_ENBL.Enable[cnt][1]);
	}

	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the selected input to the clock distribution chip.
		// Note: The default input is the on-board synthesizer.
		M377Devices.ClockDist.Payload[9][1] = 0x02;

	// Hardware reset the clock distribution chip.
	AD9512Write(p_HW, M377_CLK_DIST, 0x0000, 0x30);
	AD9512Write(p_HW, M377_CLK_DIST, 0x005A, 0x01);
	AD9512Write(p_HW, M377_CLK_DIST, 0x0000, 0x10);
	AD9512Write(p_HW, M377_CLK_DIST, 0x005A, 0x01);
	// Configure the clock distribution chip by loading the default values.
	for (cnt = 0; cnt < 22; cnt++)
		AD9512Write(p_HW, M377_CLK_DIST, M377Devices.ClockDist.Payload[cnt][0], M377Devices.ClockDist.Payload[cnt][1]);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}

// Intitialize the Model 378 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M378Initialize(s_Handle *p_HW, s_M378Devices M378Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Query the hardware for the number of receiver and transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 378 RX channels process a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Reset the clock distribution chip.
	LMK01000Write(p_HW, M378_CLK_DIST, 0x0, 0x8000010);
	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) >= 2)
		// No external clock was detected, so change the clock distribution chip input to CLKin1.
		// Note: The default input is the external clock (CLKin0).
		M378Devices.ClockDist.Payload[9][1] &= 0xDFFFFFF;
	// Configure the clock distribution chip by loading the modified default values.
	for (cnt = 0; cnt < 10; cnt++)
		LMK01000Write(p_HW, M378_CLK_DIST, M378Devices.ClockDist.Payload[cnt][0], M378Devices.ClockDist.Payload[cnt][1]);
	// Synchronize the outputs of the clock distribution chip.
	LMK01000Sync(p_HW, M378_CLK_DIST_SYNC);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Reset the quad ADC SPI port that is assigned to RX channels #1 through #4.
	AD9653Write(p_HW, M378_RX1TO4_ADC, 0x000, 0x38);
	// Reset the quad ADC SPI port that is assigned to RX channels #5 through #8.
	AD9653Write(p_HW, M378_RX5TO8_ADC, 0x000, 0x38);

	// Configure each quad ADC by loading the default values.
	//	for (cnt = 0; cnt < 24; cnt++)
	for (cnt = 0; cnt < 23; cnt++)
	{
		// Configure the quad ADC that is assigned to RX channels #1 through #4.
		AD9653Write(p_HW, M378_RX1TO4_ADC, M378Devices.RX1TO4_ADC.Payload[cnt][0], M378Devices.RX1TO4_ADC.Payload[cnt][1]);
		// Configure the quad ADC that is assigned to RX channels #5 through #8.
		AD9653Write(p_HW, M378_RX5TO8_ADC, M378Devices.RX5TO8_ADC.Payload[cnt][0], M378Devices.RX5TO8_ADC.Payload[cnt][1]);
	}

	// Activate reset on the quad ADC that is assigned to RX channels #1 through #4.
	AD9653Write(p_HW, M378_RX1TO4_ADC, 0x008, 0x03);
	// Activate reset on the quad ADC that is assigned to RX channels #5 through #8.
	AD9653Write(p_HW, M378_RX5TO8_ADC, 0x008, 0x03);
	// Deactivate reset on the quad ADC that is assigned to RX channels #1 through #4.
	AD9653Write(p_HW, M378_RX1TO4_ADC, 0x008, 0x00);
	// Deactivate reset on the quad ADC that is assigned to RX channels #5 through #8.
	AD9653Write(p_HW, M378_RX5TO8_ADC, 0x008, 0x00);
	// Wait 145 ms for the ADC to complete initialization (2.9 Msamples @ 20 Msps).
	MicrosecondTimer(145000);
	
	// Reset the trim DACs.
	// Note: Trim DACs are only present on DC coupled units.
	AD5628Write(p_HW, M378_RX1TO4_TRIM_DAC, 0x7, 0x0, 0x0);
	AD5628Write(p_HW, M378_RX5TO8_TRIM_DAC, 0x7, 0x0, 0x0);
	// Enable the internal reference of the trim DACs.
	// Note: Trim DACs are only present on DC coupled units.
	AD5628Write(p_HW, M378_RX1TO4_TRIM_DAC, 0x8, 0x0, 0x1);
	AD5628Write(p_HW, M378_RX5TO8_TRIM_DAC, 0x8, 0x0, 0x1);
	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 8; cnt++)
	{
		AD5628Write(p_HW, M378_RX1TO4_TRIM_DAC, 0x3, M378Devices.RX1TO4_TRIM_DAC.Payload[cnt][0], M378Devices.RX1TO4_TRIM_DAC.Payload[cnt][1]);
		AD5628Write(p_HW, M378_RX5TO8_TRIM_DAC, 0x3, M378Devices.RX5TO8_TRIM_DAC.Payload[cnt][0], M378Devices.RX5TO8_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the RX amplifier enables from the default structure values.
	// Note: Amplifiers are only present on RX DC coupled units.
	ADA4927Write(p_HW, CMD_STAT1, M378Devices.RX1TO2_AMP_ENBL.Enable[0], M378Devices.RX1TO2_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M378Devices.RX3TO4_AMP_ENBL.Enable[0], M378Devices.RX3TO4_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M378Devices.RX5TO6_AMP_ENBL.Enable[0], M378Devices.RX5TO6_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M378Devices.RX7TO8_AMP_ENBL.Enable[0], M378Devices.RX7TO8_AMP_ENBL.Enable[1]);
	
	// Phase align the clocks to valid data.
	return SerDesInit(p_HW);
}
