/****************************************************  Utility Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are specific to the SifFPGA application example.

	Rev History:  03-11-2019:	Some function and variable names changed for consistency across products.
				  11-28-2018:	Fixed incorrect BAR reference in ClearLatency() plus style and format changes.
				  08-08-2018:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Utility Functions  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <errno.h>

#include "driver_functions.h"
#include "sigfpga.h"
#include "global_functions.h"
#include "channel_functions.h"
#include "dma_functions.h"

// Clear the PCI latency performance measurement.
void ClearLatency(s_Handle *p_HW)
{
	// Write to the PCI latency address.
	p_HW->DEV.BAR2[PCI_LATENCY] = 0x0ull;
}

// Create a file of sine wave samples equivalent to the size of a DMA page.
// Return of non-zero indicates a failure opening the sine wave text file or a warning that the samples do not end on zero phase.
int DataFile(char *p_File, unsigned PageSize, double FsDivisor, unsigned DataItemSize)
{
	FILE *fp;
	unsigned data_items_count = 0;
	double phase = 0;
	double phase_select = 0;

	// Open the requested file and write enough samples to fill a single DMA page.
	if((fp = fopen(p_File, "w")) == NULL)
		return errno;
	else
	{
		// Store enough samples to fill the number of 64-bit records required to complete a DMA page.
		for(data_items_count = 0; data_items_count < (PageSize / 8) * (64 / DataItemSize); data_items_count++)
		{
			// Compute the amplitude of the DAC sample based on phase and write it as the current data item to the payload record.
			fprintf(fp, "%d\n", (int)((pow(2, (double)DataItemSize - 1) - 1) * sin(2 * 3.141592654 * phase / FsDivisor)));

			// Compute the phase of the next sample.
			phase = fmod(++phase_select, FsDivisor);
			// Restart the phase select counter if the phase is zero.
			if (!phase) phase_select = 0;
		}
		// Close the file.
		fclose(fp);
	}
	
	// Return a warning if the ending phase is not zero.
	if (phase)
		return PHASE_BREAK;
	else
		return 0;
}

// Preload a data buffer with data samples from a file.
// Return non-zero indicates a failure opening the sample data text file to read.
int DataLoad(char *p_File, void **p_Address, unsigned PageSize, unsigned NumPages, unsigned DataItemSize)
{
	FILE *fp;
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	unsigned page_count = 0;
	unsigned record_count = 0;
	unsigned data_items_count = 0;
	int sample = 0;
	unsigned long long *virtual_address;

	// Sequence through the number of pages requested.
	for(page_count=0; page_count < NumPages; page_count++)
	{
		// Open the requested file and read enough samples to fill a single DMA page.
		if((fp = fopen(p_File, "r")) == NULL)
			return errno;
		else
		{
			// Get the virtual address of the first 64-bit record in the current page.
			virtual_address = (unsigned long long *)p_Address[page_count];
	
			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < PageSize / 8; record_count++)
			{
				// Clear the current 64-bit record.
				virtual_address[record_count] = 0x0ull;
				
				for(data_items_count = 0; data_items_count < 64 / DataItemSize; data_items_count++)
				{
					// Read the next sample from the file.
					if (fscanf(fp, "%d", &sample) != 1)
						return EIO;
					
					// Write the sample as the current data item to the payload record.
					virtual_address[record_count] |= ((max >> (64-DataItemSize)) & (unsigned long long)sample) << data_items_count*DataItemSize;
				}
			}
		// Close the file.
		fclose(fp);
		}
	}
	return OK;
}

// Write the contents of the data buffer to the specified file from the first page to the number of pages requested.
// Return of non-zero indicates indicates a failure opening the output data text file to write.
int DataSave(char *p_File, void **p_Address, unsigned PageSize, unsigned NumPages, unsigned DataItemSize)
{
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	unsigned page_count = 0;
	unsigned data_items_count = 0;
	unsigned record_count = 0;
	int sample = 0;
	unsigned long long *virtual_address;

	// Open the requested file and process samples from the data buffer.
	if((fp = fopen(p_File, "w")) == NULL)
		return errno;
	else
	{
		// Sequence through all of the number of pages requested.
		for(page_count=0; page_count < NumPages; page_count++)
		{
			// Get the virtual address of the first 64-bit record in each page.
			virtual_address = (unsigned long long *)p_Address[page_count];

			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < PageSize / 8; record_count++)

				// Sequence through every sample in the 64-bit record.
				for (data_items_count = 0; data_items_count < 64 / DataItemSize; data_items_count++)
					{
						// Extract the next sample value from the 64-bit record.
						sample = (int)(virtual_address[record_count] >> (data_items_count * DataItemSize)) & (max >> (64 - DataItemSize));
						// Sign extend negative numbers.
						if (sample & (one << (DataItemSize - 1))) sample |= (max << DataItemSize);
						// Print the sample value to the file.
						fprintf(fp, "%d\n", sample);
					}
		}
		// Close the file.
		fclose(fp);
	}
	return OK;
}

// Free all resources allocated to a device.
// Return of non-zero indicates an error occurred closing the device or releasing memory resources.
int FreeResources(s_Handle *p_HW, unsigned ChannelCount, s_UIOScan UIO)
{
	int status = OK;
	unsigned chancnt;
	unsigned index = 0;
	s_User *User = NULL;
	
	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Free memory allocated to the error flags for each channel.
	if (User->Error)
		for (chancnt = 1; chancnt <= ChannelCount; chancnt++)
			free(User->Error[chancnt-1]);
	free(User->Error);
	
	// Free memory allocated to the data buffer attached to the User structure.
	if (User->p_DataBuffer)
		for(chancnt = 1; chancnt < ChannelCount; chancnt++)
			if (User->p_DataBuffer[chancnt-1])
			{
				for (index = 0; index < p_HW->DMA->PagesPerMark; index++)
					free(User->p_DataBuffer[chancnt-1][index]);
				free(User->p_DataBuffer[chancnt-1]);
			}
	free(User->p_DataBuffer);
		
	// Close the DMA devices.
	for (index = 0; index < ChannelCount; index++)
		if (p_HW->DMA[index].fd >= 0)
			status = DMAClose(p_HW->DMA[index].fd);
	
	// Free memory allocated to the DMA structure attached to the hardware handle.
	free(p_HW->DMA);

	// Close the Adapter device.
	if (p_HW->DEV.fd >= 0)
		status = DEVClose(UIO, &p_HW->DEV);
	
	// Free the hardware handle.
	free(p_HW);

	return status;
}

// Report the response time in units of microseconds for the last interrupt that was serviced.
// Note: The interrupt timeout (see below) will be reported if shorter than the actual response time.
float InterruptResponse(s_Handle *p_HW)
{
	// Read the interrupt latency address.
	return (float)0.004 * (float)p_HW->DEV.BAR0[INTERRUPT_LATENCY];
}

// Set the interrupt timeout in increments of 4 ns to a maximum value of (2^28)-1, or 1.074 sec.
void InterruptTimeout(s_Handle *p_HW, unsigned IntTimeout)
{
	// Write to the interrupt timeout address.
	p_HW->DEV.BAR0[INTERRUPT_TIMEOUT] = (unsigned long long)IntTimeout;
}

// Report the size of the QDR II+ SRAM address.
// Returns the number of bits.
unsigned MemSize(s_Handle *p_HW)
{
	// Read the identification address.
	return (unsigned)((p_HW->DEV.BAR2[IDENTIFICATION] >> 52) & 0xFF);
}

// Command the device to perform a test of every available address in the selected SRAM.
// SRAMSelect = 1:  SRAM-A
// SRAMSelect = 2:  SRAM-B
// SRAMSelect = 3:  SRAM-A and SRAM-B simultaneously
// Return of zero indicates success, non-zero indicates the number of errors recorded.
int MemTest(s_Handle *p_HW, unsigned SRAMSelect)
{
	unsigned status = 0;
	unsigned long long value;
	unsigned watchdog = 0;


	// Write to the SRAM command/status address to start the test on the selected SRAM.
	p_HW->DEV.BAR2[SRAM_CMD_STAT] = SRAMSelect;

	// Wait for the test to complete by monitoring the busy status.
	do
	{
		// Pause while the firmware exercises every address in the SRAM.
		MicrosecondTimer(15000);
		// Read the SRAM command/status address to monitor the status.
		value = p_HW->DEV.BAR2[SRAM_CMD_STAT];
		// Increment the watchdog counter to avoid getting caught in this loop.
		watchdog++;
	}
	while (value && watchdog != 10);

	// Read the SRAM error register to retrieve the number of errors recorded.
	value = p_HW->DEV.BAR2[SRAM_ERROR];

	// Return a unique value if the watchdog timer expired.
	if (watchdog == 10)
		status = TIME_OUT;
	// Return the recorded error count for SRAM-A if selected.
	else if (SRAMSelect == 1)
		status = value & 0xFFFFFFFFull;
	// Return the recorded error count for SRAM-B if selected.
	else if (SRAMSelect == 2)
		status = (value >> 32) & 0xFFFFFFFFull;
	// Return the sum of the recorded error counts for SRAM-A and SRAM-B if selected.
	else if (SRAMSelect == 3)
		status = (value & 0xFFFFFFFFull) + ((value >> 32) & 0xFFFFFFFFull);

	return status;
}

// Benchmark the maximum attainable PCI performance of RX and/or TX DMA transactions.
// WARNING: DMA buffers must be allocated prior to executing this function or the computer will likely crash!
void PciBenchmark(s_Handle *p_HW)
{
	unsigned long long value = 0;
	unsigned long long int_mask = 0;
	unsigned busy = 0;

	// Read the global interrupt mask addrss to record the current settings.
	int_mask = p_HW->DEV.BAR0[GLOBAL_INT_MASK];
	// Write to the global interrupt mask address to disable interrupts during the benchmark.
	p_HW->DEV.BAR0[GLOBAL_INT_MASK] = 0x0;

	// Write to the PCI throughput address to select specific channels (8:0) and initiate transactions (9).
	p_HW->DEV.BAR2[PCI_THROUGHPUT] = 0x1FFull;

	// Pause while DMA transactions are counted for a quarter second.
	MicrosecondTimer(300000);

	// Write to the PCI throughput address to discontinue all DMA transfers.
	p_HW->DEV.BAR2[PCI_THROUGHPUT] = 0x0ull;

	// Monitor the status of all channels with a polling loop and terminate when all channels stop processing.
	do
	{
		// Check the current operating state of each channel to determine if any are still active.
		value = p_HW->DEV.BAR0[DMA_1TO2_STATUS];
		busy = (value | (value >> 32)) & 0x00005000ull;
		value = p_HW->DEV.BAR0[DMA_3TO4_STATUS];
		busy |= (value | (value >> 32)) & 0x00005000ull;
		value = p_HW->DEV.BAR0[DMA_5TO6_STATUS];
		busy |= (value | (value >> 32)) & 0x00005000ull;
		value = p_HW->DEV.BAR0[DMA_7TO8_STATUS];
		busy |= (value | (value >> 32)) & 0x00005000ull;
	}
	while (busy);

	// Write to the global interrupt mask address to restore the original state.
	p_HW->DEV.BAR0[GLOBAL_INT_MASK] = int_mask;
}

// Report the measured instantaneous PCI throughput for the last 250 ms and the maximum DMA latency since the last clear.
void PciPerformance(s_Handle *p_HW, float *p_Performance)
{
	unsigned long long value = 0;

	// Read the PCI throughput address to retrieve the last measurement.
	// Note: A new PCI throughput measurement is stored every 250 ms.
	value = p_HW->DEV.BAR2[PCI_THROUGHPUT];

	// Record the measured receiver throughput in Gbytes/sec.
	p_Performance[0] = (float)(value & 0x00000000FFFFFFFFull) / (float)1e9;
	// Record the measured transmitter throughput in Gbytes/sec.
	p_Performance[1] = (float)((value >> 32) & 0x00000000FFFFFFFFull) / (float)1e9;

	// Read the PCI latency address to retrieve the last measurement.
	value = p_HW->DEV.BAR2[PCI_LATENCY];

	// Record the measured receiver latency in units of micrseconds.
	p_Performance[2] = (float)(value & 0x00000000FFFFFFFFull) * (float)4e-3;
	// Record the measured transmitter latency in units of micrseconds.
	p_Performance[3] = (float)((value >> 32) & 0x00000000FFFFFFFFull) * (float)4e-3;
}

// Query each channel for any status change and update the user defined structure of the hardware handle accordingly.
void ProcessChannels(s_Handle *p_HW)
{
	unsigned chancnt = 0;		// Variable to index channel numbers in loops.
	unsigned pagecnt = 0;		// Counter to sequence through the DMA pages of a chapter.
	s_User *User = NULL;		// Pointer to the user defined structure of the hardware handle.
	s_ChannelStatus Channel;	// Structure to store the results returned by a channel status query to the device.
	s_DMAStatus DMA;			// Structure to store the results returned by a DMA status query to the device.
	s_GlobalStatus Global;		// Structure to store the results returned by an interrupt status query to the device.

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Determine which channels have a status update.
	GlobalStatus(p_HW, &Global);

	// Update the user defined structure of any channel reporting a status change.
	for (chancnt = 1; chancnt <= User->RXChannelCount + User->TXChannelCount; chancnt++)
	{
		if (Global.Channel[chancnt-1])
		{
			// Determine which channel status events require an update.
			ChannelStatus(p_HW, chancnt, &Channel);

			// Increment the converter (ADC/DAC) error count if an error is detected.
			if (Channel.Error[0]) User->Error[chancnt-1][0]++;

			// Increment the FIFO error count if an error is detected.
			if (Channel.Error[1]) User->Error[chancnt-1][1]++;

			// Process the DMA buffer contents if the DMA refresh marker is detected.
			if (Channel.DMAMarker)
			{
				// Read the DMA controller status to get the number of the current page that the device is servicing.
				DMAStatus(p_HW, chancnt, &DMA);

				// Unwind the roll-over condition that occurs when the last DMA page is completed and the index starts back at zero.
				if (DMA.CurrentPage <= p_HW->DMA[chancnt-1].PageNumber)
					DMA.CurrentPage += p_HW->DMA[chancnt-1].PagesPerMark;

				// Post a lost data error if the current page is more than two markers beyond the last page processed.
				// Note: This technique cannot detect loss of data if the DMA engine completes a full cycle through all of the available pages before the test is performed.
				if (DMA.CurrentPage > p_HW->DMA[chancnt-1].PageNumber + (2 * p_HW->DMA[chancnt-1].PagesPerMark))
					User->Error[chancnt-1][2]++;

				// Copy the contents of the last chapter of DMA pages that were refreshed between DMA markers into data buffers accessible to the main application.
				// Note: This assumes that there was only one DMA marker update since the last check.
				for (pagecnt = 0; pagecnt < p_HW->DMA[chancnt-1].PagesPerMark; pagecnt++)
				{
					// Map the virtual address of the next DMA page.
					if (DMAMap(&p_HW->DMA[chancnt-1]))
					{
						User->Error[chancnt-1][3]++;
						break;
					}
				
					// Determine whether the current channel is a transmitter or receiver.
					if (chancnt > User->RXChannelCount)
						// Copy from the data buffer to the DMA buffer if this is a transmitter channel.
						memcpy(p_HW->DMA[chancnt-1].Address, User->p_DataBuffer[chancnt-1][pagecnt], p_HW->DMA[chancnt-1].BurstSizeB * p_HW->DMA[chancnt-1].BurstCount);
					else
						// Copy from the DMA buffer to the data buffer if this is a receiver channel.
						memcpy(User->p_DataBuffer[chancnt-1][pagecnt], p_HW->DMA[chancnt-1].Address, p_HW->DMA[chancnt-1].BurstSizeB * p_HW->DMA[chancnt-1].BurstCount);

					// Unmap the current DMA page.
					if (DMAUnmap(&p_HW->DMA[chancnt-1]))
						User->Error[chancnt-1][3]++;
				}
			}
		}
	}
}

// Initializes all varibles in the User structure to default values.
void UserInit(s_Handle *p_HW, s_User *p_User)
{
	// Point the device handle to the user structure.
	p_HW->User = p_User;
	// Set all of the size and count variables to zero.
	p_User->RXChannelCount = 0;
	p_User->RXDataItemSize = 0;
	p_User->TXChannelCount = 0;
	p_User->TXDataItemSize = 0;
	// Set all User structure pointers to NULL.
	p_User->Error = NULL;
	p_User->p_DataBuffer = NULL;
}